import numpy as np
import cv2 as cv

from vkit.image.type import (
    VImage,
    VImageKind,
)
from vkit.label.type import (
    VImageScoreMap,
    VImageMask,
)
from .type import VImageGrid


def create_image_from_image_grid(image_grid: VImageGrid, image_kind: VImageKind):
    ndim = VImageKind.to_ndim(image_kind)
    if ndim == 2:
        shape = (image_grid.image_height, image_grid.image_width)
    elif ndim == 3:
        num_channels = VImageKind.to_num_channels(image_kind)
        assert num_channels
        shape = (image_grid.image_height, image_grid.image_width, num_channels)
    else:
        raise NotImplementedError()

    dtype = VImageKind.to_dtype(image_kind)
    mat = np.zeros(shape, dtype=dtype)
    return VImage(mat=mat, kind=image_kind)


def create_image_score_map_from_image_grid(image_grid):
    shape = (image_grid.image_height, image_grid.image_width)
    mat = np.zeros(shape, dtype=np.float32)
    return VImageScoreMap(mat=mat)


def create_image_mask_from_image_grid(image_grid):
    shape = (image_grid.image_height, image_grid.image_width)
    mat = np.zeros(shape, dtype=np.uint8)
    return VImageMask(mat=mat)


def blend_src_to_dst_image(
    src_image: VImage,
    src_image_grid: VImageGrid,
    dst_image_grid: VImageGrid,
):
    map_y, map_x = src_image_grid.generate_remap_params(dst_image_grid)
    dst_image_mat = cv.remap(src_image.mat, map_x, map_y, cv.INTER_LINEAR)
    return VImage(mat=dst_image_mat, kind=src_image.kind)


def blend_src_to_dst_image_score_map(
    src_image_score_map: VImageScoreMap,
    src_image_grid: VImageGrid,
    dst_image_grid: VImageGrid,
):
    map_y, map_x = src_image_grid.generate_remap_params(dst_image_grid)
    dst_image_score_map_mat = cv.remap(src_image_score_map.mat, map_x, map_y, cv.INTER_LINEAR)
    return VImageScoreMap(mat=dst_image_score_map_mat)


def blend_src_to_dst_image_mask(
    src_image_mask: VImageMask,
    src_image_grid: VImageGrid,
    dst_image_grid: VImageGrid,
):
    map_y, map_x = src_image_grid.generate_remap_params(dst_image_grid)
    dst_image_mask_mat = cv.remap(src_image_mask.mat, map_x, map_y, cv.INTER_LINEAR)
    return VImageMask(mat=dst_image_mask_mat)


def debug():
    from vkit.opt import get_data_folder
    folder = get_data_folder(__file__)

    from vkit.label.type import VPoint
    from vkit.label.visualization import visualize_points

    src_image = VImage.from_file(f'{folder}/Lenna.png')
    dst_image = src_image.clone()
    dst_image.mat.fill(255)

    dst_image.to_file(f'{folder}/dst.png')

    # Mock grid.
    from .type import VImageGrid

    ys = list(range(0, src_image.height, src_image.height // 25))
    if ys[-1] != src_image.height - 1:
        ys.append(src_image.height - 1)
    xs = list(range(0, src_image.width, src_image.width // 25))
    if xs[-1] != src_image.width - 1:
        xs.append(src_image.width - 1)

    points_2d = []
    for y in ys:
        points = []
        for x in xs:
            points.append(VPoint(y=y, x=x))
        points_2d.append(points)
    src_image_grid = VImageGrid(points_2d=points_2d, grid_size=10)

    import itertools

    visualize_points(src_image, itertools.chain.from_iterable(src_image_grid.points_2d
                                                              )).to_file(f'{folder}/src-grid.png')

    # # Check identity.
    # dst_image = blend_src_to_dst_image(src_image, src_image_grid, src_image_grid)
    # # assert (dst_image.mat == src_image.mat).all()
    # print(
    #     'identity ratio',
    #     (dst_image.mat == src_image.mat).sum() / (src_image.height * src_image.width * 3),
    # )
    # dst_image.to_file(f'{folder}/identity.png')

    # Random grid.
    import random

    # for idx, y in enumerate(ys):
    #     if idx == 0 or idx == len(ys) - 1:
    #         continue
    #     if random.random() < 0.5:
    #         ys[idx] = y + src_image.height // 50
    #     else:
    #         ys[idx] = y - src_image.height // 50

    # for idx, x in enumerate(xs):
    #     if idx == 0 or idx == len(xs) - 1:
    #         continue
    #     if random.random() < 0.5:
    #         xs[idx] = x + src_image.width // 50
    #     else:
    #         xs[idx] = x - src_image.width // 50

    # dst_image_grid = VImageGrid(points_2d=[])
    # for y in ys:
    #     points = []
    #     for x in xs:
    #         points.append(VPoint(y=y, x=x))
    #     dst_image_grid.points_2d.append(points)

    points_2d = []
    for y in ys:
        points = []
        for x in xs:
            if 0 < y < src_image.height - 1:
                if random.random() < 0.5:
                    ry = y + src_image.height // 150
                else:
                    ry = y - src_image.height // 150
            else:
                ry = y

            if 0 < x < src_image.width - 1:
                if random.random() < 0.5:
                    rx = x + src_image.width // 150
                else:
                    rx = x - src_image.width // 150
            else:
                rx = x
            points.append(VPoint(y=ry, x=rx))

        points_2d.append(points)

    dst_image_grid = VImageGrid(points_2d=points_2d)

    dst_image = blend_src_to_dst_image(src_image, src_image_grid, dst_image_grid)
    dst_image.to_file(f'{folder}/random.png')

    dst_image = blend_src_to_dst_image(src_image, src_image_grid, dst_image_grid)
    visualize_points(
        dst_image,
        itertools.chain.from_iterable(dst_image_grid.points_2d),
    ).to_file(f'{folder}/random-grid.png')


def debug_cv():
    dst_polygon_points = np.array(
        [
            [51., 90.],
            [60., 90.],
            [61., 100.],
            [51., 100.],
        ],
        dtype=np.float32,
    )

    src_polygon_points = np.array(
        [
            [50., 90.],
            [60., 90.],
            [60., 100.],
            [50., 100.],
        ],
        dtype=np.float32,
    )
    trans_mat = cv.getPerspectiveTransform(
        dst_polygon_points,
        src_polygon_points,
        # cv.DECOMP_SVD,
    )
    print(trans_mat)
