from typing import Iterable, Optional, Tuple, Union

from PIL import ImageDraw, ImageColor
import numpy as np
import cv2 as cv

from vkit.image.type import VImage
from .type import (
    VPoint,
    VLine,
    VBox,
    VPointList,
    VImageMask,
    VImageScoreMap,
    VPolygon,
)

# https://mokole.com/palette.html
_DISTINCT_HEX_RGB_COLORS = [
    # darkgreen
    '#006400',
    # darkblue
    '#00008b',
    # maroon3
    '#b03060',
    # red
    '#ff0000',
    # yellow
    '#ffff00',
    # burlywood
    '#deb887',
    # lime
    '#00ff00',
    # aqua
    '#00ffff',
    # fuchsia
    '#ff00ff',
    # cornflower
    '#6495ed',
]


def get_rgb_tuple(hex_color) -> Tuple[int, int, int]:
    # https://pillow.readthedocs.io/en/stable/reference/ImageColor.html#color-names
    return ImageColor.getrgb(hex_color)  # type: ignore


def get_complementary_rgb_or_rgba_color(color):
    complementary_color = (255 - color[0], 255 - color[1], 255 - color[2])
    if len(color) == 4:
        complementary_color = (*complementary_color, color[-1])
    return complementary_color


def get_color(
    hex_color: Optional[str],
    distinct_idx: Optional[int] = None,
    alpha: Optional[int] = None,
):
    if hex_color == 'distinct':
        assert distinct_idx is not None
        hex_color = _DISTINCT_HEX_RGB_COLORS[distinct_idx % len(_DISTINCT_HEX_RGB_COLORS)]

    rgb = get_rgb_tuple(hex_color)
    if alpha:
        # RGBA.
        return (*rgb, alpha)
    else:
        # RGB.
        return rgb


def visualize_points(
    image: VImage,
    points: Union[VPointList, Iterable[VPoint]],
    style: str = 'circle-1',
    fill: str = 'distinct',
):
    pil_image = image.to_rgb_image().to_pil_image()
    draw = ImageDraw.Draw(pil_image)

    if style == 'pixel':
        for idx, point in enumerate(points):
            draw.point([(point.x, point.y)], fill=get_color(fill, idx))

    elif style.startswith('circle-'):
        radius = int(style[len('circle-'):])
        for idx, point in enumerate(points):
            up = max(point.y - radius, 0)
            down = min(point.y + radius, image.height - 1)
            left = max(point.x - radius, 0)
            right = min(point.x + radius, image.width - 1)
            draw.ellipse((left, up, right, down), fill=get_color(fill, idx))

    else:
        raise NotImplementedError()

    return VImage.from_pil_image(pil_image)


def visualize_lines(
    image: VImage,
    lines: Iterable[VLine],
    fill: str = 'distinct',
    show_index: bool = False,
):
    pil_image = image.to_rgb_image().to_pil_image()
    draw = ImageDraw.Draw(pil_image)

    for idx, line in enumerate(lines):
        color = get_color(fill, idx)
        draw.line(line.to_xy_pairs(), fill=color)
        if show_index:
            draw.text(line.to_center_point().to_xy_pair(), str(idx), fill=color)

    return VImage.from_pil_image(pil_image)


def visualize_polygons(
    image: VImage,
    polygons: Iterable[VPolygon],
    fill: Optional[str] = 'distinct',
    outline: Optional[str] = 'distinct',
    alpha: int = 128,
    show_index: bool = False,
    index_color: str = 'distinct',
):
    pil_image = image.to_rgba_image().to_pil_image()
    pil_polygon_image = VImage.from_shape((*image.shape, 4), fill_value=0).to_pil_image()
    draw = ImageDraw.Draw(pil_polygon_image)

    for idx, polygon in enumerate(polygons):
        fill_color = None
        outline_color = None

        if fill:
            fill_color = get_color(fill, idx, alpha)
            outline_color = fill_color

        if outline:
            outline_color = get_color(outline, idx, alpha if fill else 255)

        assert fill_color or outline_color
        draw.polygon(tuple(polygon.to_flatten_xy_pairs()), fill=fill_color, outline=outline_color)

        if show_index:
            text_fill = fill_color or outline_color
            assert text_fill and index_color

            if index_color == 'distinct':
                text_fill = get_complementary_rgb_or_rgba_color(text_fill)
            else:
                text_fill = get_color(index_color, idx, alpha if fill else 255)

            y_mean = round(np.mean([point.y for point in polygon.points]))  # type: ignore
            x_mean = round(np.mean([point.x for point in polygon.points]))  # type: ignore

            draw.text((x_mean, y_mean), str(idx), fill=text_fill)

    pil_image.alpha_composite(pil_polygon_image)
    return VImage.from_pil_image(pil_image)


def visualize_boxes(
    image: VImage,
    boxes: Iterable[VBox],
    fill: Optional[str] = 'distinct',
    outline: Optional[str] = 'distinct',
    alpha: int = 128,
    show_index: bool = False,
    index_color: str = 'distinct',
):
    return visualize_polygons(
        image=image,
        polygons=(box.to_polygon() for box in boxes),
        fill=fill,
        outline=outline,
        alpha=alpha,
        show_index=show_index,
        index_color=index_color,
    )


def visualize_image_score_map(image_score_map: VImageScoreMap):
    mat = image_score_map.mat.copy()

    # Rescale to [0, 255]
    val_min = np.min(mat)
    mat -= val_min
    val_max = np.max(mat)
    mat *= 255.0
    mat /= val_max

    mat = np.clip(mat, 0, 255).astype(np.uint8)

    # Apply color map.
    color_mat = cv.applyColorMap(mat, cv.COLORMAP_JET)
    color_mat = cv.cvtColor(color_mat, cv.COLOR_BGR2RGB)

    return VImage(mat=color_mat)


def visualize_scale_image_score_map(image_score_map: VImageScoreMap, image_mask: VImageMask):
    # (0, +)
    mat = image_score_map.mat.copy()

    # Scale to 127.5.
    mat *= 127.5

    # Clip [0, 255].
    mat = np.clip(mat, 0, 255)
    # To grayscale.
    mat = mat.astype(np.uint8)

    # Apply color map.
    color_mat = cv.applyColorMap(mat, cv.COLORMAP_JET)
    color_mat = cv.cvtColor(color_mat, cv.COLOR_BGR2RGB)

    # Set background as black.
    color_mat[image_mask.mat == 0] = 0

    return VImage(mat=color_mat)


def visualize_image_mask(image_mask: VImageMask):
    mat = image_mask.mat

    new_mat = np.zeros_like(mat)
    np.putmask(new_mat, mat > 0, 255)

    return VImage(mat=new_mat)


def blend_image_with_image_mask(
    foreground_image: VImage,
    foreground_image_mask: VImageMask,
    background_image: VImage,
    foreground_alpha: float = 0.5,
):
    fg_mask_mat = (foreground_image_mask.mat > 0)
    fg_mask_mat = fg_mask_mat.astype(np.float32) * foreground_alpha
    bg_mask_mask = 1.0 - fg_mask_mat

    fg_img_mat = foreground_image.mat.astype(np.float32) / 255.0
    bg_img_mat = background_image.mat.astype(np.float32) / 255.0

    mat = cv.addWeighted(
        fg_img_mat * np.expand_dims(fg_mask_mat, -1),
        255.0,
        bg_img_mat * np.expand_dims(bg_mask_mask, -1),
        255.0,
        0.0,
    )
    mat = mat.astype(np.uint8)
    return VImage(mat=mat)


def debug():
    from vkit.opt import get_data_folder
    folder = get_data_folder(__file__)

    visualize_points(
        image=VImage.from_shape((100, 100)),
        points=[VPoint(y=0, x=0), VPoint(y=50, x=50),
                VPoint(y=20, x=20)],
    ).to_file(f'{folder}/visualize_points-default.png')

    visualize_points(
        image=VImage.from_shape((100, 100)),
        points=[VPoint(y=0, x=0), VPoint(y=50, x=50),
                VPoint(y=20, x=20)],
        fill='red',
    ).to_file(f'{folder}/visualize_points-style-pixel-fill-red.png')

    visualize_lines(
        image=VImage.from_shape((100, 100)),
        lines=[
            VLine.from_flatten_xy_pairs([0, 0, 20, 99]),
            VLine.from_flatten_xy_pairs([50, 50, 50, 99]),
            VLine.from_flatten_xy_pairs([50, 50, 99, 50]),
        ]
    ).to_file(f'{folder}/visualize_lines-default.png')

    visualize_lines(
        image=VImage.from_shape((100, 100)),
        lines=[
            VLine.from_flatten_xy_pairs([0, 0, 20, 99]),
            VLine.from_flatten_xy_pairs([50, 50, 50, 99]),
            VLine.from_flatten_xy_pairs([50, 50, 99, 50]),
        ],
        show_index=True,
    ).to_file(f'{folder}/visualize_lines-show_index.png')

    visualize_polygons(
        image=VImage.from_file(f'{folder}/Lenna.png'),
        polygons=[
            VPolygon.from_xy_pairs([
                (10, 10),
                (100, 10),
                (100, 30),
                (10, 30),
            ]),
            VPolygon.from_xy_pairs([
                (110, 110),
                (200, 110),
                (200, 130),
                (110, 200),
            ]),
        ],
    ).to_file(f'{folder}/visualize_polygons-default.png')

    visualize_polygons(
        image=VImage.from_file(f'{folder}/Lenna.png'),
        polygons=[
            VPolygon.from_xy_pairs([
                (10, 10),
                (100, 10),
                (100, 30),
                (10, 30),
            ]),
            VPolygon.from_xy_pairs([
                (110, 110),
                (200, 110),
                (200, 130),
                (110, 200),
            ]),
        ],
        fill=None,
        outline='red'
    ).to_file(f'{folder}/visualize_polygons-fill-None-outline-red.png')

    visualize_polygons(
        image=VImage.from_file(f'{folder}/Lenna.png'),
        polygons=[
            VPolygon.from_xy_pairs([
                (10, 10),
                (100, 10),
                (100, 30),
                (10, 30),
            ]),
            VPolygon.from_xy_pairs([
                (110, 110),
                (200, 110),
                (200, 130),
                (110, 200),
            ]),
        ],
        show_index=True,
    ).to_file(f'{folder}/visualize_polygons-with-index.png')
